/******************************************************************************
 *
 * AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: 
 * Simon Vey, Thomas Witkowski, Andreas Naumann, Simon Praetorius, et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * This file is part of AMDiS
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/



/** \file CouplingProblemStat.h */

#ifndef AMDIS_COUPLING_PROBLEM_STAT_H
#define AMDIS_COUPLING_PROBLEM_STAT_H

#include <vector>
#include <set>
#include <list>
#include "AMDiS_fwd.h"
#include "ProblemStat.h"
#include "Initfile.h"
#include <boost/lexical_cast.hpp>

namespace AMDiS {

  namespace detail
  {

    /** \brief
    * This class defines a coupled stationary problem definition in sequential
    * computations. 
    */
    template<typename ProblemStatType>
    class CouplingProblemStat
    {
    public:
      /// Constructor
      CouplingProblemStat(std::string name_)
	: name(name_),
	  nComponents(0),
	  nMeshes(0),
	  refinementManager(NULL),
	  coarseningManager(NULL)
      {}

      /// Destructor
      virtual ~CouplingProblemStat() {}

      /// add problem by number
      virtual void addProblem(ProblemStatType* prob)
      {
	problems.push_back(prob);
	nComponents += prob->getNumComponents();
      };

      /// Initialisation of the problem.
      virtual void initialize(Flag initFlag,
			      ProblemStatSeq *adoptProblem = NULL,
			      Flag adoptFlag = INIT_NOTHING)
      { FUNCNAME("CouplingProblemStat::initialize()");

	// create one refinement-/coarseningmanager for all problems
	if (refinementManager != NULL && coarseningManager != NULL) { 
	  WARNING("refinement-/coarseningmanager already created\n");
	} else {
	  if (!adoptProblem)
	    createRefCoarseManager();
	  else {
	    refinementManager = adoptProblem->getRefinementManager();
	    coarseningManager = adoptProblem->getCoarseningManager();
	  }
	}
    
	if (refinementManager == NULL || coarseningManager == NULL) 
	  WARNING("no refinement-/coarseningmanager created\n");

	// create Meshes and FeSpaces


	// all problems must have the same dimension (?)
	int dim = 0;
	Parameters::get(name + "->dim", dim);
	TEST_EXIT(dim)("No problem dimension specified for \"%s->dim\"!\n",
		      name.c_str());

	std::map<std::string, Mesh*> meshByName;
	std::vector<std::set<Mesh*> > meshesForProblems(problems.size());
	std::map<std::pair<Mesh*, int>, FiniteElemSpace*> feSpaceMap;
	for (size_t i = 0; i < problems.size(); ++i) {
	  TEST_EXIT(problems[i])("problem[%d] does not exist!\n",i);
	  for (size_t j = 0; j < problems[i]->getNumComponents(); j++) {
	    // mesh

	    std::string meshName("");
	    Parameters::get(problems[i]->getName() + "->mesh", meshName);
	    TEST_EXIT(meshName != "")("No mesh name specified for \"%s->mesh\"!\n",
				      problems[i]->getName().c_str());

	    if (meshByName.find(meshName) == meshByName.end()) {
	      Mesh *newMesh = new Mesh(meshName, dim);
	      meshByName[meshName] = newMesh;
	      meshes.push_back(newMesh);
	      meshesForProblems[i].insert(newMesh);
	      nMeshes++;
	    } else
	      meshesForProblems[i].insert(meshByName[meshName]);

	    problems[i]->setComponentMesh(j, meshByName[meshName]);
	      
	    // feSpace
	    int degree = 1;
	    Parameters::get(problems[i]->getName() + "->polynomial degree[" + 
			    boost::lexical_cast<std::string>(j) + "]", degree);

	    if (feSpaceMap[std::pair<Mesh*, int>(meshByName[meshName], degree)] == NULL) {
	      std::stringstream s;
	      s << problems[i]->getName() << "->feSpace[" << j << "]";
      
	      FiniteElemSpace *newFeSpace = 
		FiniteElemSpace::provideFeSpace(NULL, Lagrange::getLagrange(dim, degree),
						meshByName[meshName], s.str());
	      feSpaceMap[std::pair<Mesh*, int>(meshByName[meshName], degree)] = newFeSpace;
	      feSpaces.push_back(newFeSpace);
	    }
      
  //           problems[i]->setComponentSpace(j, feSpaceMap[std::pair<Mesh*, int>(meshByName[meshName], degree)]);

	  }
	}

	for (size_t i = 0; i < problems.size(); i++) {
	  std::vector<Mesh*> problemMeshes(meshesForProblems[i].begin(), meshesForProblems[i].end());
	  problems[i]->setMeshes(problemMeshes);
	  problems[i]->setRefinementManager(refinementManager);
	  problems[i]->setCoarseningManager(coarseningManager);
	  problems[i]->initialize(INIT_ALL - INIT_MESH);
	}

	for (size_t i = 0; i < meshes.size(); i++) {
	  int globalRefinements = 0;

	  // If AMDiS is compiled for parallel computations, the global refinements are
	  // ignored here. Later, each rank will add the global refinements to its
	  // private mesh.
  #ifndef HAVE_PARALLEL_DOMAIN_AMDIS
	  Parameters::get(meshes[i]->getName() + "->global refinements",
			  globalRefinements);
  #endif

	  bool initMesh = initFlag.isSet(INIT_MESH);

	  // Initialize the meshes if there is no serialization file.
	  if (initMesh && meshes[i] && !(meshes[i]->isInitialized())) {
	    meshes[i]->initialize();
	    refinementManager->globalRefine(meshes[i], globalRefinements);
	  }
	}
      }

      void createRefCoarseManager() 
      {
	FUNCNAME("CouplingProblemStat::createRefCoarseManager()");
    
	int dim = 0;
	Parameters::get(name + "->dim", dim);
	TEST_EXIT(dim)("No problem dimension specified for \"%s->dim\"!\n",
		      name.c_str());

	switch (dim) {
	case 1:
	  coarseningManager = new CoarseningManager1d();
	  refinementManager = new RefinementManager1d();
	  break;
	case 2:
	  coarseningManager = new CoarseningManager2d();
	  refinementManager = new RefinementManager2d();
	  break;
	case 3:
	  coarseningManager = new CoarseningManager3d();
	  refinementManager = new RefinementManager3d();
	  break;
	default:
	  ERROR_EXIT("invalid dim!\n");
	}
      }

      /// Returns number of managed problems
      virtual int getNumProblems() 
      { 
	return problems.size(); 
      }

      /// Implementation of ProblemStatBase::getNumComponents()
      virtual int getNumComponents() 
      { 
	return nComponents; 
      }

      /** \brief
      * Returns the problem with the given number. If only one problem
      * is managed by this master problem, the number hasn't to be given.
      */
      virtual ProblemStatType *getProblem(int number = 0)
      { 
	return problems[number]; 
      }

      /// Returns \ref meshes[i]
      inline Mesh* getMesh(int number = 0) 
      {
	return meshes[number]; 
      }

      /// Returns \ref meshes
      inline std::vector<Mesh*> getMeshes() 
      {
	return meshes; 
      }

      /// Returns \ref refinementManager.
      inline RefinementManager* getRefinementManager(int comp = 0) 
      { 
	return refinementManager; 
      }

      /// Returns \ref refinementManager.
      inline CoarseningManager* getCoarseningManager(int comp = 0) 
      { 
	return coarseningManager; 
      }

      /// Returns the name of the problem
      inline virtual std::string getName() 
      { 
	return name; 
      }

    protected:
      
      /// Name of this problem.
      std::string name;

      /// Number of problem components
      int nComponents;

      /** \brief
      * Number of problem meshes. If all components are defined on the same mesh,
      * this number is 1. Otherwise, this variable is the number of different meshes
      * within the problem.
      */
      int nMeshes;

      /// FE spaces of this problem.
      std::vector<FiniteElemSpace*> feSpaces;

      /// Meshes of this problem.
      std::vector<Mesh*> meshes;

      /** \brief
      * All actions of mesh refinement are performed by refinementManager.
      * If new refinement algorithms should be realized, one has to override
      * RefinementManager and give one instance of it to AdaptStationary.
      */
      RefinementManager *refinementManager;

      /** \brief
      * All actions of mesh coarsening are performed by coarseningManager.
      * If new coarsening algorithms should be realized, one has to override
      * CoarseningManager and give one instance of it to AdaptStationary.
      */
      CoarseningManager *coarseningManager;

      std::vector<ProblemStatType*> problems;
    };

  } // end namespace detail

  typedef detail::CouplingProblemStat<ProblemStat> CouplingProblemStat;
  
} // end namespace AMDiS

#endif
