/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/


#ifndef EXTENSIONS_REFINEMENT_EXPRESSION_H
#define EXTENSIONS_REFINEMENT_EXPRESSION_H

#include "ProblemStat.h"
#include "ElementFunction.h"
#include "RefineFunctions.h"
#include "RefineOperations.h"
#include "Expressions.h"

namespace AMDiS { namespace extensions {


/** \brief
 * Base class for Refinement structure to perform local anisotropic refinement
 */
class RefinementExpression
{
public:

  RefinementExpression(ProblemStat& prob, int component = 0, bool doCalcMeshSize_ = false)
    : mesh(prob.getMesh(component)), 
      adaptInfo(NULL),
      refineOperation(NULL),
      numRefinements0(15),
      doCalcMeshSize(doCalcMeshSize_),
      onlyRefine(false),
      finished(true),
      refinementManagerCreated(false)
  {
    coarseningManager = prob.getCoarseningManager();
    refinementManager = prob.getRefinementManager();
    numRefinements = numRefinements0;    
    refineOperation = new StandardRefineOperation;
  }
  
  RefinementExpression(Mesh *mesh_, bool doCalcMeshSize_ = false)
    : mesh(mesh_), 
      adaptInfo(NULL),
      refineOperation(NULL),
      numRefinements0(15),
      doCalcMeshSize(doCalcMeshSize_),
      onlyRefine(false),
      finished(true),
      refinementManagerCreated(true)
  {	
    FUNCNAME("RefinementExpression::RefinementLevel()");
    
    switch (mesh->getDim()) {
    case 1:
      coarseningManager = new CoarseningManager1d();
      refinementManager = new RefinementManager1d();
      break;
    case 2:
      coarseningManager = new CoarseningManager2d();
      refinementManager = new RefinementManager2d();
      break;
    case 3:
      coarseningManager = new CoarseningManager3d();
      refinementManager = new RefinementManager3d();
      break;
    default:
      ERROR_EXIT("invalid dim!\n");
    }

    numRefinements = numRefinements0;
    
    refineOperation = new StandardRefineOperation;
  }
  
  virtual ~RefinementExpression() 
  {
    finalize();
    
    if (refinementManagerCreated) {
      delete coarseningManager;
      delete refinementManager;
    }
    
    if (refineOperation)
      delete refineOperation;
  }
  
  void finalize()
  {
    if (!finished) {
      #ifdef HAVE_PARALLEL_DOMAIN_AMDIS
      Parallel::MeshDistributor::globalMeshDistributor->checkMeshChange();
      #endif
    }
    finished = true;
  }
  
  template <typename Term>
  bool refine(Term term) 
  {
    FUNCNAME("RefinementExpression::refine()");

    double minH = 0.0, maxH = 1.0;
    int minLevel = 100, maxLevel = 0;
    int oldMeshChangeIdx = mesh->getChangeIndex();
    
    bool meshChanged = true;
    Flag markFlag;
    int oldNr = 0, oldOldNr = 0, i = 0;
    while (meshChanged && i < numRefinements) {
      markElements(term, markFlag);
      meshChanged = refineMesh(markFlag, onlyRefine);
      
      // compare nr of vertices
      int nr = mesh->getNumberOfVertices();
      meshChanged = meshChanged && oldOldNr!=nr && oldNr!=nr;
      
      if (meshChanged) {
	if (doCalcMeshSize) {
	  calcMeshSizes(minH, maxH, minLevel, maxLevel, false); 
	  MSG("(local) mesh sizes: [%f, %f], Vs: %d, ELs: %d\n", 
	    minH, maxH, nr, mesh->getNumberOfElements());
	} else {
	  MSG("(local) #Vertics: %d, #Elements: %d\n", nr, mesh->getNumberOfElements());
	}
      }
			
      oldOldNr = oldNr;
      oldNr = nr; 
      i++;
    }
    
    calcMeshSizes(minH, maxH, minLevel, maxLevel, true); 
    MSG("Final (global) mesh: [%f, %f], Vs: %d, ELs: %d, Level: [%d, %d]\n",
      minH, maxH, mesh->getNumberOfVertices(), mesh->getNumberOfElements(), minLevel, maxLevel);
      
    return oldMeshChangeIdx != mesh->getChangeIndex();
  }
  

  template <typename Term>
  bool refine(int numRefinements_, Term term, bool onlyRefine_= false) 
  { 
    numRefinements = numRefinements_; 
    setOnlyRefine(onlyRefine_);
    bool refined = refine(term);  
    numRefinements = numRefinements0; 
    return refined;
  }
  

  template <int N, typename Term>
  bool refine(Term term, bool onlyRefine_= false) 
  { 
    numRefinements = N; 
    setOnlyRefine(onlyRefine_);
    bool refined = refine(term);  
    numRefinements = numRefinements0; 
    return refined;
  }
  
  
  template<typename Term>
  void markElements(Term term, Flag &markFlag) 
  {    
    bool elMarkRefine = false, elMarkCoarsen = false;
    
    GenericOperatorTerm<Term> ot(term);
    
    int dim = mesh->getDim();
    Quadrature* quad = Quadrature::provideQuadrature(dim, 0); // 1 point in the center

    Flag traverseFlag = Mesh::CALL_LEAF_EL | Mesh::FILL_COORDS | Mesh::FILL_GRD_LAMBDA;
    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(mesh, -1, traverseFlag);
    while (elInfo) {
      term.initElement(&ot, elInfo, NULL, quad);
      
      int refineLevel = boost::numeric_cast<int>(term(0)); // eval expression
      int oldLevel = elInfo->getLevel();
      elInfo->getElement()->setMark( calcMark(refineLevel, oldLevel) );

      elMarkRefine |= elInfo->getElement()->getMark() == 1;
      elMarkCoarsen |= elInfo->getElement()->getMark() == -1;
      elInfo = stack.traverseNext(elInfo);
    }
      
    markFlag= 0;
    if(elMarkRefine) 
      markFlag = 1;
    if(elMarkCoarsen) 
      markFlag|= 2;
  }

  int getNumRefinements() 
  {
    return numRefinements;
  }

  void calcMeshSizes(double& minH, double& maxH, int& minLevel, int& maxLevel, bool allReduce = false) 
  {
    FixVec<WorldVector<double>, VERTEX> coords(mesh->getDim(), NO_INIT);

    TraverseStack stack;
    ElInfo *elInfo = stack.traverseFirst(mesh, -1, Mesh::CALL_LEAF_EL | Mesh::FILL_COORDS);
    minH = 1e15; maxH = 0.0;
    int k = 0;
    minLevel = 100;
    maxLevel = 0;
    while (elInfo) {
      maxLevel = std::max(maxLevel,elInfo->getLevel());
      minLevel = std::min(minLevel,elInfo->getLevel());
      coords = elInfo->getCoords();
      double h = 0.0;
      for (int i = 0; i < coords.size(); i++) {
        for (int j = 0; j < coords.size(); j++) {
	  if (i != j)
            h = std::max(h,norm(coords[i]-coords[j]));
        }
      }
      minH = std::min(h, minH);
      maxH = std::max(h, maxH);
      elInfo = stack.traverseNext(elInfo);
      k++;
    }
    minLevel += mesh->getMacroElementLevel();
    maxLevel += mesh->getMacroElementLevel();
    
#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
    if (allReduce) {
      Parallel::mpi::globalMin(minH);
      Parallel::mpi::globalMax(maxH);
      Parallel::mpi::globalMin(minLevel);
      Parallel::mpi::globalMax(maxLevel);
    }
#endif
  }


  double calcMeshSize(ElInfo *elInfo, bool allReduce = false) 
  {
    FixVec<WorldVector<double>, VERTEX> coords(mesh->getDim(), NO_INIT);
    coords = elInfo->getCoords();
    double h = 0.0;
    for (int i = 0; i < coords.size(); i++) {
      for (int j = 0; j < coords.size(); j++) {
        if (i != j)
          h = std::max(h,norm(coords[i]-coords[j]));
      }
    }
#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
    if (allReduce)
      Parallel::mpi::globalMax(h);
#endif

    return h;
  }


  int calcMark(double refineH, double currentH)
  {
    return (refineH < currentH ? 
        1 : (refineH > currentH * (mesh->getDim() == 1 ? 
          2.0 : (mesh->getDim() == 2 ? 
          std::sqrt(2.0) : 
          std::sqrt(2.0)/2.0 + 0.5)) ? 
        -1 : 
        0));
  }


  int calcMark(int refineLevel, int currentLevel)
  {
    int levelDiff = refineLevel - currentLevel;
    return (levelDiff > 0 ? 1 : (levelDiff < 0 ? -1 : 0));
  }

	
  bool refineMesh(Flag markFlag, bool onlyRefine) 
  {
    finished = false;
    
    int oldSize = mesh->getNumberOfVertices();
    refineOperation->beforeRefine(adaptInfo, markFlag);
    if (markFlag.isSet(1))
      refinementManager->refineMesh(mesh);
      
    refineOperation->beforeCoarsen(adaptInfo, markFlag);
    if (markFlag.isSet(2) && !onlyRefine)
      coarseningManager->coarsenMesh(mesh);
      
    refineOperation->afterCoarsen(adaptInfo, markFlag);
    
    if (markFlag.isSet(1) || markFlag.isSet(2)) {
      int newSize = mesh->getNumberOfVertices();
      if (oldSize != newSize) 
        return true;
    }
    return false;
  }
  
  void setRefineOperation(AdaptInfo* adaptInfo_,
			  StandardRefineOperation* refineOperation_)
  {
    if (refineOperation)
      delete refineOperation;
    
    adaptInfo = adaptInfo_;
    refineOperation = refineOperation_;
  }

  void setOnlyRefine(bool onlyRefine_)
  {
    onlyRefine = onlyRefine_;
  }

  void setCalcMeshSize(bool calc) { doCalcMeshSize = calc; }
  
  RefinementManager* getRefinementManager() { return refinementManager; }
  CoarseningManager* getCoarseningManager() { return coarseningManager; }

protected:

  Mesh* mesh;
  RefinementManager* refinementManager;
  CoarseningManager* coarseningManager;
  
  AdaptInfo* adaptInfo;
  StandardRefineOperation* refineOperation;

  int numRefinements;
  int numRefinements0;
  
  bool doCalcMeshSize;
  bool onlyRefine;
  bool finished;
  bool refinementManagerCreated;
};

} }

#endif // EXTENSIONS_REFINEMENT_EXPRESSION_H
