/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/
#include "NavierStokes_TH_MultiPhase_RB.h"

namespace AMDiS { namespace base_problems {
  
using namespace std;

NavierStokes_TH_MultiPhase_RB::NavierStokes_TH_MultiPhase_RB(const std::string &name_, bool createProblem) :
  super(name_, createProblem),
  viscosity1(-1.0),
  viscosity2(1.0),
  density1(1.0),
  density2(1.0),
  densityPhase(NULL),
  viscosityPhase(NULL),
  multiPhase(NULL)
{ FUNCNAME("NavierStokes_TH_MultiPhase_RB::_constructor()");

  Initfile::get(name + "->viscosity1", viscosity1); // viscosity of fluid 1
  Initfile::get(name + "->viscosity2", viscosity2); // viscosity of fluid 2
  Initfile::get(name + "->density1", density1); // density of fluid 1
  Initfile::get(name + "->density2", density2); // density of fluid 2

  if (viscosity1 <= 0.0 && viscosity > 0.0)
    viscosity1 = viscosity;

  TEST_EXIT(viscosity1 > 0.0)("Positive viscosity is necessary!\n");
}


NavierStokes_TH_MultiPhase_RB::~NavierStokes_TH_MultiPhase_RB()
{
  if (multiPhase) { delete multiPhase; multiPhase = NULL; }
  if (densityPhase) { delete densityPhase; densityPhase = NULL; }
  if (viscosityPhase) { delete viscosityPhase; viscosityPhase = NULL; }
}


void NavierStokes_TH_MultiPhase_RB::initData()
{
  densityPhase = new DOFVector<double>(prob->getFeSpace(0), "densityPhase");
  viscosityPhase = new DOFVector<double>(prob->getFeSpace(0), "viscosityPhase");
  
  densityPhase->set(density1);
  viscosityPhase->set(viscosity1);

  super::initData();
}


void NavierStokes_TH_MultiPhase_RB::initTimestep(AdaptInfo *adaptInfo)
{
  super::initTimestep(adaptInfo);

  if (multiPhase) {
    LinearInterpolation1 dLI(density1, density2);
    LinearInterpolation1 vLI(viscosity1, viscosity2);
    transformDOF(multiPhase, densityPhase, &dLI);
    transformDOF(multiPhase, viscosityPhase, &vLI);
  }
}


void NavierStokes_TH_MultiPhase_RB::fillOperators()
{
  using namespace extensions;
  int degree_u = prob->getFeSpace(0)->getBasisFcts()->getDegree();
//   int degree_p = prob->getFeSpace(dow)->getBasisFcts()->getDegree();
  
  WorldVector<DOFVector<double>* > vel;
  for (unsigned i = 0; i < dow; i++)
    vel[i]= prob->getSolution()->getDOFVector(i);

  WorldVector<DOFVector<double>*> U;
  WorldVector<DOFVector<double>*> U_old;
  for (unsigned i = 0; i < dow; ++i) {
    U[i] = prob->getStageSolution(i);
    U_old[i] = prob->getUnVec(i);
  }
    
  // fill operators for prob
  for (size_t i = 0; i < dow; ++i) {
    /// < d_t(u_i) , psi >
    Operator *opTime = new Operator(getFeSpace(i), getFeSpace(i));
    if (density1 == density2 || !densityPhase)
      opTime->addTerm(new Simple_ZOT(density1));  
    else
      opTime->addTerm(new VecAtQP_ZOT(densityPhase));
    prob->addMatrixOperator(opTime, i, i, prob->getInvTauGamma(), prob->getInvTauGamma());

    Operator *opTimeOld = new Operator(getFeSpace(i));
    if (density1 == density2 || !densityPhase)
      opTimeOld->addTerm(new VecAtQP_ZOT(prob->getTimeRhsVec(i), new AMDiS::Factor<double>(density1, degree_u)));  
    else
      opTimeOld->addTerm(new Vec2AtQP_ZOT(densityPhase, prob->getTimeRhsVec(i)));
    prob->addVectorOperator(opTimeOld, i);

    /// < u*grad(u_i) , psi >
    Operator *opUGradU = new Operator(getFeSpace(i), getFeSpace(i));
    if (density1 == density2 || !densityPhase)
        opUGradU->addTerm(new WorldVec_FOT(U, -density1), GRD_PHI);
    else
        opUGradU->addTerm(new WorldVecPhase_FOT(densityPhase, U, -1.0), GRD_PHI);
    opUGradU->setUhOld(U[i]);
    prob->addVectorOperator(*opUGradU, i);
      
    /// < u^old*grad(u'_i) , psi >
    for (size_t j = 0; j < dow; ++j) {
      Operator *opUGradV = new Operator(getFeSpace(i),getFeSpace(i));
      if (density1 == density2 || !densityPhase)
	opUGradV->addTerm(new VecAndPartialDerivative_FOT(U_old[j], j, density1), GRD_PHI);
      else
	opUGradV->addTerm(new Vec2AndPartialDerivative_FOT(densityPhase, U_old[j], j), GRD_PHI);
      prob->addMatrixOperator(*opUGradV, i, i);
    }

    for (size_t j = 0; j < dow; ++j) {
      Operator *opVGradU = new Operator(getFeSpace(i), getFeSpace(j));      
      if (density1 == density2 || !densityPhase) 
	opVGradU->addTerm(new PartialDerivative_ZOT(U_old[i], j, density1));
      else
	opVGradU->addTerm(new VecAndPartialDerivative_ZOT(densityPhase, U_old[i], j));
      prob->addMatrixOperator(*opVGradU, i, j);
    }

    /// Diffusion-Operator (including Stress-Tensor for space-dependent viscosity
    addLaplaceTerm(i);

    /// < p , d_i(psi) >
    Operator *opGradP = new Operator(getFeSpace(i),getFeSpace(dow));
    opGradP->addTerm(new PartialDerivative_FOT(i, -1.0), GRD_PSI);
    opGradP->setUhOld(prob->getStageSolution(dow));
    prob->addVectorOperator(*opGradP, i);
    prob->addMatrixOperator(*opGradP, i, dow, &minus1, &minus1);
  
    /// external force, i.e. gravitational force
    if (std::abs(force[i]) > DBL_TOL) {
      Operator *opForce = new Operator(getFeSpace(i), getFeSpace(i));
      if (density1 == density2 || !densityPhase)
	opForce->addTerm(new Simple_ZOT(density1*force[i]));
      else
	opForce->addTerm(new VecAtQP_ZOT(densityPhase, NULL, force[i]));
      prob->addVectorOperator(*opForce, i);
    }
  }

  /// div(u) = 0
  for (size_t i = 0; i < dow; ++i) {
    /// < d_i(u'_i) , psi >
    Operator *opDivU = new Operator(getFeSpace(dow),getFeSpace(i));
    opDivU->addTerm(new PartialDerivative_FOT(i), GRD_PHI);
    opDivU->setUhOld(U[i]);
    prob->addVectorOperator(*opDivU, dow);
    prob->addMatrixOperator(*opDivU, dow, i, &minus1, &minus1);
  }
  
  
  Operator *opZero = new Operator(getFeSpace(dow),getFeSpace(dow));
  opZero->addTerm(new Simple_ZOT(0.0));
  prob->addMatrixOperator(*opZero, dow, dow);
}


void NavierStokes_TH_MultiPhase_RB::addLaplaceTerm(int i)
{
  using namespace extensions;
  /// < alpha*[grad(u)+grad(u)^t] , grad(psi) >
  if (viscosity1 != viscosity2 && viscosityPhase) {
    for (size_t j = 0; j < dow; ++j) {
      Operator *opLaplaceUi1 = new Operator(getFeSpace(i), getFeSpace(j));
      opLaplaceUi1->addTerm(new VecAtQP_IJ_SOT(viscosityPhase, NULL, j, i));
      opLaplaceUi1->setUhOld(prob->getStageSolution(j));
      prob->addVectorOperator(*opLaplaceUi1, i, &minus1, &minus1);
      prob->addMatrixOperator(*opLaplaceUi1, i, j);
    }
  }
  
  /// < alpha*grad(u'_i) , grad(psi) >
  Operator *opLaplaceUi = new Operator(getFeSpace(i), getFeSpace(i));
  if (viscosity1 == viscosity2 || !viscosityPhase) {
    opLaplaceUi->addTerm(new Simple_SOT(viscosity1));
  } else {
    opLaplaceUi->addTerm(new VecAtQP_SOT(viscosityPhase));
  }
  opLaplaceUi->setUhOld(prob->getStageSolution(i));
  prob->addVectorOperator(*opLaplaceUi, i, &minus1, &minus1);
  prob->addMatrixOperator(*opLaplaceUi, i, i);
}

} }
