/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/

#ifndef NAVIER_STOKES_TAYLOR_HOOD_MULTIPHASE_H
#define NAVIER_STOKES_TAYLOR_HOOD_MULTIPHASE_H

#include "NavierStokes_TaylorHood_RB.h"

namespace AMDiS { namespace base_problems {

/** \ingroup NavierStokes_TaylorHood
 * \brief
 * Navier-Stokes multi-phase problem, using Taylor Hood elements
 * Standard implementation for two phases, but can be extended to
 * many more, by defining the multiPhase variable and overloading
 * the initTimestep, where densityPhase and viscosityPhase are defined
 * depending on the phases
 */
class NavierStokes_TH_MultiPhase_RB : public NavierStokes_TaylorHood_RB
{
public: // typedefs

  typedef NavierStokes_TaylorHood_RB super;

public: // methods

  NavierStokes_TH_MultiPhase_RB(const std::string &name_, bool createProblem = true);

  /// deletes multiPhase, viscosityPhase and densityPhase
  ~NavierStokes_TH_MultiPhase_RB();

  /** definition of constant multiPhase and const
   * viscosity/density DOFVectors,
   * initTimeInterface of super
   **/
  void initData() override;

  /** initTimestep of super and
   * initialization of densityPhase and viscosityPhase
   **/
  void initTimestep(AdaptInfo *adaptInfo) override;

  /** pointer to the multiPhase DOFVector, that
   * indicates the different phases. Will be initialized
   * in \ref initTimeInterface with const 1
   **/
  DOFVector<double> *getMultiPhase() { return multiPhase; }

  void fillOperators() override;
  void addLaplaceTerm(int i) override;
  
  // get-methods
  
  DOFVector<double>* getViscosityPhase()
  {
    return viscosityPhase;
  }
  
  DOFVector<double>* getDensityPhase()
  {
    return densityPhase;
  }
  
  double getViscosity(int i=0)
  {
    if (i == 0)
      return viscosity1;
    else
      return viscosity2;
  }
  
  double getDensity(int i=0)
  {
    if (i == 0)
      return density1;
    else
      return density2;
  }
  
protected: // variables

  ///viscosity of phase 1
  double viscosity1;
  
  ///viscosity of phase 2
  double viscosity2;
  
  ///density of phase 1
  double density1;
  
  ///density of phase 2
  double density2;

  /// phase dependent density
  DOFVector<double> *densityPhase;
  
  /// phase dependent viscosity
  DOFVector<double> *viscosityPhase;
  
  /// phase inticator
  DOFVector<double> *multiPhase;

};


/** linear interpolation between two values (like density, viscosity)
 *  using a phase-field variable in [-1,1]
 **/
class LinearInterpolation1 : public AbstractFunction<double, double>
{
public:

  LinearInterpolation1(double val1_, double val2_) :
    AbstractFunction<double, double>(1),
    val1(val1_), val2(val2_) {}

  double operator()(const double &phase) const { 
    return 0.5*(phase+1.0)*val1 - 0.5*(phase-1.0)*val2;
  }

private:
  
  double val1;
  double val2;
};

/** linear interpolation between two values (like density, viscosity)
 *  using a phase-field variable in [0,1]
 **/
class LinearInterpolation0 : public AbstractFunction<double, double>
{
public:

  LinearInterpolation0(double val1_, double val2_) :
    AbstractFunction<double, double>(1),
    val1(val1_), val2(val2_) {}

  double operator()(const double &phase) const { 
    return phase*val1 + (1.0-phase)*val2;
  }

private:
  
  double val1;
  double val2;
};

} }

#endif // NAVIER_STOKES_TAYLOR_HOOD_MULTIPHASE_H

