/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/

#include "Helpers.h"
#include "POperators.h"

namespace AMDiS { namespace base_problems {
  
namespace detail {

template<typename P> 
NavierStokes_TaylorHood<P>::NavierStokes_TaylorHood(const std::string &name_, bool createProblem) :
  super(name_, createProblem),
  velocity(NULL),
  laplaceType(0),
  nonLinTerm(2),
  oldTimestep(0.0),
  viscosity(1.0),
  density(1.0),
  minus1(-1.0),
  theta(0.5+1.e-2),
  theta1(1-theta),
  minusTheta1(-theta1),
  fileWriter(NULL)
{  
  // fluid viscosity
  Initfile::get(name_ + "->viscosity", viscosity);
  // fluid density
  Initfile::get(name_ + "->density", density);

  // type of laplace operator: 
  //   0... div(nu*grad(u)), 
  //   1... div(0.5*nu*(grad(u)+grad(u)^T))
  Initfile::get(name_ + "->laplace operator", laplaceType); 

  // type of non-linear term: 
  //   0... u^old*grad(u_i^old), 
  //   1... u*grad(u_i^old), 
  //   2... u^old*grad(u_i)
  //   3... (1)+(2)-(0)
  Initfile::get(name_ + "->non-linear term", nonLinTerm); 

  // theta-scheme parameter
  Initfile::get(name_ + "->theta", theta); 
  theta1 = 1.0-theta;
  minusTheta1 = -theta1;

  // volume force
  force.set(0.0);
  Initfile::get(name_ + "->force", force);

  oldSolution.resize(self::dow);
  for (size_t i = 0; i < self::dow; i++)
    oldSolution[i] = NULL;
}

template<typename P> 
NavierStokes_TaylorHood<P>::~NavierStokes_TaylorHood() 
{   
  if (velocity != NULL) {
    delete velocity;
    velocity = NULL;
  }
  
  for (size_t i = 0; i < self::dow; i++) {
    if (oldSolution[i] != NULL)
      delete oldSolution[i];
    oldSolution[i] = NULL;
  }
 
  if (fileWriter) { 
    delete fileWriter;
    fileWriter = NULL;
  }
}


template<typename P> 
void NavierStokes_TaylorHood<P>::initData()
{ 
  if (velocity == NULL)
    velocity = new DOFVector<WorldVector<double> >(self::getFeSpace(0), "velocity");
  
  for (size_t i = 0; i < self::dow; i++)
    oldSolution[i] = new DOFVector<double>(self::getFeSpace(i), "old(v_"+extensions::Helpers::toString(i)+")");
  
#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
  for (size_t i = 0; i < self::dow; i++)
    Parallel::MeshDistributor::globalMeshDistributor->addInterchangeVector(oldSolution[i]);
#endif
  
  fileWriter = new FileVectorWriter(self::name + "->velocity->output", self::getFeSpace()->getMesh(), velocity);

  super::initData();
}


template<typename P> 
void NavierStokes_TaylorHood<P>::transferInitialSolution(AdaptInfo *adaptInfo)
{ 
  calcVelocity();
  for (size_t i = 0; i < self::dow; i++)
    oldSolution[i]->copy(*self::prob->getSolution()->getDOFVector(i));
  
  super::transferInitialSolution(adaptInfo);
}


template<typename P> 
void NavierStokes_TaylorHood<P>::fillOperators()
{ 
  using namespace extensions;
  WorldVector<DOFVector<double>* > vel;
  for (size_t k = 0; k < self::dow; k++)
    vel[k] = self::prob->getSolution()->getDOFVector(k);
  
  const FiniteElemSpace* feSpace_u = self::getFeSpace(0);
  const FiniteElemSpace* feSpace_p = self::getFeSpace(self::dow);
  
  // fill operators for prob
  for (size_t i = 0; i < self::dow; ++i) {
    /// < (1/tau)*u'_i , psi >
    Operator *opTime = new Operator(feSpace_u, feSpace_u);
    addZOT(opTime, density);
    self::prob->addMatrixOperator(*opTime, i, i, self::getInvTau(), self::getInvTau());
    
    /// < (1/tau)*u_i^old , psi >
    Operator *opTimeOld = new Operator(feSpace_u, feSpace_u);
    addZOT(opTimeOld, valueOf(getOldSolution(i)) * density);
    self::prob->addVectorOperator(*opTimeOld, i, self::getInvTau(), self::getInvTau());

    /// < u^old*grad(u_i^old) , psi >
    if (nonLinTerm == 0 || nonLinTerm == 3) {
      Operator *opUGradU0 = new Operator(feSpace_u, feSpace_u);
      addFOT(opUGradU0, valueOf(vel) * density, GRD_PHI);
      opUGradU0->setUhOld(self::prob->getSolution()->getDOFVector(i));
      if (nonLinTerm == 0)
        self::prob->addVectorOperator(*opUGradU0, i);
      else
        self::prob->addVectorOperator(*opUGradU0, i, &minus1, &minus1);
    }
    
    /// < u*grad(u_i^old) , psi >
    if (nonLinTerm == 1 || nonLinTerm == 3) {
      for (size_t k = 0; k < self::dow; ++k) {
        Operator *opUGradU1 = new Operator(feSpace_u, feSpace_u);
	addZOT(opUGradU1, derivativeOf(vel[i], k) * density);
        self::prob->addMatrixOperator(*opUGradU1, i, k);
      }
    }

    /// < u^old*grad(u_i) , psi >
    if (nonLinTerm == 2 || nonLinTerm == 3) {
      for (size_t k = 0; k < self::dow; ++k) {
        Operator *opUGradU2 = new Operator(feSpace_u, feSpace_u);
	addFOT(opUGradU2, valueOf(vel[k]) * density, k, GRD_PHI);
        self::prob->addMatrixOperator(*opUGradU2, i, i);
      }
    }

    /// Diffusion-Operator
    addLaplaceTerm(i);
  
    /// < p , d_i(psi) >
    Operator *opGradP = new Operator(feSpace_u, feSpace_p);
    addFOT(opGradP, -1.0, i, GRD_PSI);
    self::prob->addMatrixOperator(*opGradP, i, self::dow);
  
    /// external force, i.e. gravitational force
    if (std::abs(force[i]) > DBL_TOL) {
      Operator *opForce = new Operator(feSpace_u, feSpace_u);
      addZOT(opForce, (density*force[i]) );
      self::prob->addVectorOperator(*opForce, i);
    }
  }

  /// div(u) = 0
  for (size_t i = 0; i < self::dow; ++i) {
    /// < d_i(u'_i) , psi >
    Operator *opDivU = new Operator(feSpace_p, feSpace_u);
    addFOT(opDivU, 1.0, i, GRD_PHI);
    self::prob->addMatrixOperator(*opDivU, self::dow, i);
  }

  Operator *opNull = new Operator(feSpace_p, feSpace_p);
  addZOT(opNull, 0.0);
  self::prob->addMatrixOperator(*opNull, self::dow, self::dow);
}


template<typename P> 
void NavierStokes_TaylorHood<P>::addLaplaceTerm(int i)
{ 
  using namespace extensions;
  const FiniteElemSpace* feSpace_u = self::getFeSpace(0);
  
  /// < alpha*[grad(u)+grad(u)^t] , grad(psi) >
  if (self::laplaceType == 1) {
    for (size_t j = 0; j < self::dow; ++j) {
      Operator *opLaplaceUi1 = new Operator(feSpace_u, feSpace_u);
      addSOT(opLaplaceUi1, viscosity, j, i);
      self::prob->addMatrixOperator(*opLaplaceUi1, i, j, &theta, &theta);

      if (std::abs(minusTheta1) > DBL_TOL) {
	opLaplaceUi1->setUhOld(self::prob->getSolution()->getDOFVector(j));
	self::prob->addVectorOperator(*opLaplaceUi1, i, &minusTheta1, &minusTheta1);
      }
    }
  }
  
  /// < alpha*grad(u'_i) , grad(psi) >
  Operator *opLaplaceUi = new Operator(feSpace_u, feSpace_u);
  addSOT(opLaplaceUi, viscosity);
  self::prob->addMatrixOperator(*opLaplaceUi, i, i, &theta, &theta);

  if (std::abs(minusTheta1) > DBL_TOL) {
    opLaplaceUi->setUhOld(self::prob->getSolution()->getDOFVector(i));
    self::prob->addVectorOperator(*opLaplaceUi, i, &minusTheta1, &minusTheta1);
  }
}


template<typename P> 
void NavierStokes_TaylorHood<P>::closeTimestep(AdaptInfo *adaptInfo)
{
  calcVelocity();
  for (size_t i = 0; i < self::dow; i++)
    oldSolution[i]->copy(*self::prob->getSolution()->getDOFVector(i));
  
  super::closeTimestep(adaptInfo);
}


template<typename P> 
void NavierStokes_TaylorHood<P>::writeFiles(AdaptInfo *adaptInfo, bool force)
{ 
  super::writeFiles(adaptInfo, force);
  self::fileWriter->writeFiles(adaptInfo, false);
}

} // end namespace detail

} }
