
#include "Expressions.h"

namespace AMDiS { namespace base_problems {
  
  namespace detail 
  {
    template<typename P>
    PhaseFieldCrystal<P>::PhaseFieldCrystal(const std::string &name_, bool createProblem) 
      : super(name_, createProblem),
	mobility_type(0),
	r(-0.4),		// temperature deviation
	density(-0.3),		// mean density
	M0(1.0),		// mobility
	oldSolution(NULL),
	stepSolution(NULL)
    {  
      Parameters::get(name_ + "->r",r);
      Parameters::get(name_ + "->density", density);
      Parameters::get(name_ + "->M0", M0);
      Parameters::get(name_ + "->mobility type", mobility_type);
    }
    
    
    template<typename P>
    void PhaseFieldCrystal<P>::initData()
    {
      super::initData();
    
      oldSolution = new SystemVector(*self::prob->getSolution());
      stepSolution = new SystemVector(*self::prob->getSolution());
    }


    template<typename P>
    void PhaseFieldCrystal<P>::transferInitialSolution(AdaptInfo *adaptInfo)
    {
      super::transferInitialSolution(adaptInfo);
      *oldSolution = *self::prob->getSolution();
      *stepSolution = *self::prob->getSolution();
    }


    template<typename P>
    void PhaseFieldCrystal<P>::fillOperators()
    {
      const FiniteElemSpace* feSpace0 = self::prob->getFeSpace(0);
      const FiniteElemSpace* feSpace1 = self::prob->getFeSpace(1);
      const FiniteElemSpace* feSpace2 = self::prob->getFeSpace(2);
      
      DOFVector<double>* rho = stepSolution->getDOFVector(1);
      DOFVector<double>* rho_old = oldSolution->getDOFVector(1);
      
      // dt(rho) = laplace(mu) - u*grad(rho)
      // -----------------------------------
      Operator *opM11 = new Operator(feSpace1, feSpace1);
      addZOT(opM11, 1.0);
      self::prob->addMatrixOperator(opM11, 1, 1);
      
      Operator *opLM = new Operator(feSpace1, feSpace0);
      if (mobility_type == 0)
	addSOT(opLM, M0);
      else if (mobility_type == 1)
	addSOT(opLM, max(abs_(valueOf(rho) + 1.5)*M0, 1.e-5));
      else if (mobility_type == 2)
	addSOT(opLM, max(valueOf(rho)*M0, 1.e-5));
      
      self::prob->addMatrixOperator(opLM, 1, 0, self::getTau(), self::getTau()); // -laplace(mu)
      
      // . . . vectorOperators . . . . . . . . . . . . . . .
      Operator *opMold = new Operator(feSpace1);
      addZOT(opMold, valueOf(rho_old));
      self::prob->addVectorOperator(opMold, 1);
      
      // mu-2*nu-laplace(nu)-(1+r)*rho = (rho_old^3) + ExtPot - eps^2/(rho_old+0.9)
      // ----------------------------------------------------------------------
      Operator *op01 = new Operator(feSpace0, feSpace1);
      addZOT(op01, -(1.0 + ref_(&r)) - 3.0 * pow<2>(valueOf(rho)) );
      addSOT(op01, 2.0);
      self::prob->addMatrixOperator(op01, 0, 1);
      
      Operator *op00 = new Operator(feSpace0, feSpace0);
      addZOT(op00, 1.0);
      self::prob->addMatrixOperator(op00, 0, 0);
      
      Operator *op02 = new Operator(feSpace0, feSpace2);
      addSOT(op02, 1.0);
      self::prob->addMatrixOperator(op02, 0, 2);
      
      // . . . vectorOperators . . . . . . . . . . . . . . .
      Operator *op0 = new Operator(feSpace0);
      addZOT(op0, -2.0 * pow<3>(valueOf(rho)) );
      self::prob->addVectorOperator(op0, 0);
      
      // 0 = nu-laplace(rho)
      // -------------------
      Operator *op21 = new Operator(feSpace2, feSpace1);
      addSOT(op21, 1.0);
      self::prob->addMatrixOperator(op21, 2, 1); // -laplace(rho)
      
      Operator *op22 = new Operator(feSpace2, feSpace2);
      addZOT(op22, 1.0);
      self::prob->addMatrixOperator(op22, 2, 2); // nu
    }


    template<typename P>
    double PhaseFieldCrystal<P>::calcEnergy()
    {
      DOFVector<double>* rho = self::prob->getSolution(1);
      DOFVector<double>* mu = self::prob->getSolution(0);
      
      return integrate( 0.5*valueOf(rho)*valueOf(mu) - 0.25*pow<4>(valueOf(rho)) );
    }


    template<typename P>
    void PhaseFieldCrystal<P>::closeTimestep(AdaptInfo *adaptInfo)
    {
      super::closeTimestep(adaptInfo);
      acceptTimestep();
    }

  } // end namespace detail

} }
