/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/

#ifndef POLARIZATION_FIELD_H
#define POLARIZATION_FIELD_H

#include "AMDiS.h"
#include "BaseProblem.h"
#include "GenericOperatorTerm.h"

namespace AMDiS { namespace base_problems {
  
namespace detail 
{  
  /** \ingroup BaseProblems
  * \brief
  * Simulation of the relaxation of an orientation field (polarization field).
  * The evolution equation reads:
  * 
  * d_t P = alpha2*laplace(P#) - alpha4*P#
  *     P#= 1/eps * (|P|^2 - 1)*P - K*laplace(P)
  * 
  * with P=(P1, P2), P#=(P#1, P#2) and 
  * |P| = sqrt(P^T*P), [laplace(P)]_i = laplace(P_i)
  * 
  * As time-discretization a semi-implicit Euler discretization is implemented.
  * 
  * A filewriter for P is provided via the initfile parameter
  * "[name]->vectorField->output". You have to set the sup-parameter 
  * "(...)->ParaView vector format: 1" and 
  * "(...)->write vector as 3d vector: 1" to visualize the field P as 
  * vectorfield in ParaView.
  */  
  template<typename ProblemStatType>
  class PolarizationField : public BaseProblem<ProblemStatType>
  {
  public: // typedefs
    
    typedef PolarizationField<ProblemStatType> self;
    typedef BaseProblem<ProblemStatType> super;
    
  public: // methods

    /// constructor
    PolarizationField(const std::string &name_);

    /// destructor
    ~PolarizationField();

    /// initialize the vectorField and corresponding fileWriter
    virtual void initData() override;

    /// calls \ref calcVectorField and \ref super::transferInitialSolution
    virtual void transferInitialSolution(AdaptInfo *adaptInfo) override;  

    /// calls \ref calcVectorField and \ref super::closeTimestep
    virtual void closeTimestep(AdaptInfo *adaptInfo) override;
    
    /// write the solution and the vectorField
    virtual void writeFiles(AdaptInfo *adaptInfo, bool force = false) override;

    // === getting/setting methods ===

    /// return a DOFVector that describes the vectorField P
    DOFVector<WorldVector<double> >* getVectorField()
    { FUNCNAME_DBG("getVectorField()");
      TEST_EXIT_DBG(vectorField != NULL)
	("vectorField is NULL!");
      return vectorField;
    }
    
    /// return a pointer to the solution in the last timestep
    DOFVector<double> *getOldSolution(int i)
    { FUNCNAME_DBG("getOldSolution()");
      TEST_EXIT_DBG(oldSolution[i] != NULL)
	("Index out of range, or oldSolution not provided for this component!\n");
      return oldSolution[i];
    }

    /// implementation of BaseProblem::fillOperators
    virtual void fillOperators() override;
    
    /// used in fillOperators to add the term (grad(P_i), grad(psi))
    virtual void fillLaplacian();

  protected: // variables

    DOFVector<WorldVector<double> >* vectorField;

    void calcVectorField()
    {
      WorldVector<DOFVector<double>*> P;
      for (size_t i = 0; i < self::dow; i++)
	P[i] = self::prob->getSolution()->getDOFVector(i);
      *vectorField << valueOf(P);
    }

    double oldTimestep;
    double minus1;
    
    double alpha2;
    double alpha4;
    double C1;
    double C4;
    double epsilon;
    double epsInv;
    double K;

    FileVectorWriter *fileWriter;
    std::vector<DOFVector<double>*> oldSolution;
  };

} // end namespace detail

typedef detail::PolarizationField<AMDiS::ProblemStat> PolarizationField;

} }

#include "PolarizationField.hh"

#endif // POLARIZATION_FIELD_H

