/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/

#include "Helpers.h"

namespace AMDiS { namespace base_problems {

namespace detail {

template<typename P> 
PolarizationField_RB<P>::PolarizationField_RB(const std::string &name_) :
  super(name_, true),
  vectorField(NULL),
  minus1(-1.0),
  alpha2(1.0),
  alpha4(1.0),
  epsilon(0.1),
  K(1.0),
  fileWriter(NULL)
{  
  Parameters::get(self::name + "->alpha2", alpha2);
  Parameters::get(self::name + "->alpha4", alpha4);
  
  Parameters::get(self::name + "->epsilon", epsilon);
  epsInv = 1.0/epsilon;
  
  Parameters::get(self::name + "->K", K);
}

template<typename P> 
PolarizationField_RB<P>::~PolarizationField_RB() 
{   
  if (vectorField != NULL) {
    delete vectorField;
    vectorField = NULL;
  }
 
  if (fileWriter) { 
    delete fileWriter;
    fileWriter = NULL;
  }
}


template<typename P> 
void PolarizationField_RB<P>::initData()
{ 
  if (vectorField == NULL)
    vectorField = new DOFVector<WorldVector<double> >(self::getFeSpace(0), "vectorField");
  
  fileWriter = new FileVectorWriter(self::name + "->vectorField->output", self::getFeSpace()->getMesh(), vectorField);

  super::initData();
}


template<typename P> 
void PolarizationField_RB<P>::transferInitialSolution(AdaptInfo *adaptInfo)
{ 
  calcVectorField();  
  super::transferInitialSolution(adaptInfo);
}


template<typename P> 
void PolarizationField_RB<P>::fillOperators()
{ 
  WorldVector<DOFVector<double>* > vec;
  WorldVector<DOFVector<double>* > vec0;
  for (size_t k = 0; k < self::dow; k++) {
    vec[k] = self::prob->getStageSolution(k);
    vec0[k] = self::prob->getUnVec(k);
  }
  
  const FiniteElemSpace* feSpace = self::getFeSpace(0);
  
  // fill operators for component P
  for (size_t i = 0; i < self::dow; ++i) {
    /// < (1/tau)*P_i , psi >
    self::prob->addTimeOperator(i, i);	// dt(P_i)

    /// Diffusion-Operator
    Operator *opLaplace = new Operator(feSpace, feSpace);
    opLaplace->setUhOld(self::prob->getStageSolution(i+self::dow)); // P#_i
    addSOT(opLaplace, constant(-alpha2));
    addZOT(opLaplace, constant(-alpha4));
    
    self::prob->addMatrixOperator(*opLaplace, i, i+self::dow, &minus1, &minus1);
    self::prob->addVectorOperator(*opLaplace, i);
  }

  // fill operators for component P#
  for (size_t i = 0; i < self::dow; ++i) {
    /// < P# , psi >
    Operator *opM = new Operator(feSpace, feSpace);
    opM->setUhOld(self::prob->getStageSolution(i+self::dow)); // P#_i
    addZOT(opM, constant(1.0));
    
    self::prob->addMatrixOperator(*opM, i+self::dow, i+self::dow, &minus1, &minus1);
    self::prob->addVectorOperator(*opM, i+self::dow);
    
    /// < (-C1 - C4*P^2)*P , psi >
    Operator *opNonlinExpl = new Operator(feSpace, feSpace);
    Operator *opNonlinImpl = new Operator(feSpace, feSpace);
    if (self::dow == 2) {
      addZOT(opNonlinExpl, epsInv * (1.0 - pow<2>(valueOf(vec[0])) - pow<2>(valueOf(vec[1]))) );
      addZOT(opNonlinImpl, epsInv * (1.0 - pow<2>(valueOf(vec0[0])) - pow<2>(valueOf(vec0[1]))) ); // (1 - |P|^2)*dP
    } else if (self::dow == 3) {
      addZOT(opNonlinExpl, epsInv * (1.0 - pow<2>(valueOf(vec[0])) - pow<2>(valueOf(vec[1])) - pow<2>(valueOf(vec[2]))) );
      addZOT(opNonlinImpl, epsInv * (1.0 - pow<2>(valueOf(vec0[0])) - pow<2>(valueOf(vec0[1])) - pow<2>(valueOf(vec0[2]))) );
    }
    self::prob->addMatrixOperator(*opNonlinImpl, i+self::dow, i, &minus1, &minus1);
    self::prob->addVectorOperator(*opNonlinExpl, i+self::dow);
        
    for (size_t j = 0; j < self::dow; ++j) {
      Operator *opNonlinImpl2 = new Operator(feSpace, feSpace);
      addZOT(opNonlinImpl2, (-2.0*epsInv) * valueOf(vec0[i]) * valueOf(vec0[j]));
      self::prob->addMatrixOperator(*opNonlinImpl2, i+self::dow, j, &minus1, &minus1);
    }
  }
  
  fillLaplacian();

  Operator *opNull = new Operator(feSpace, feSpace);
  addZOT(opNull, constant(0.0));
  
  for (size_t i = 0; i < self::dow; ++i) {
    for (size_t j = i+1; j < self::dow; ++j) {
      self::prob->addMatrixOperator(*opNull, i, j);
      self::prob->addMatrixOperator(*opNull, j, i);
      self::prob->addMatrixOperator(*opNull, i, j+self::dow);
      self::prob->addMatrixOperator(*opNull, j, i+self::dow);
      self::prob->addMatrixOperator(*opNull, i+self::dow, j+self::dow);
      self::prob->addMatrixOperator(*opNull, j+self::dow, i+self::dow);
    }
  }
}


template<typename P> 
void PolarizationField_RB<P>::fillLaplacian()
{
  const FiniteElemSpace* feSpace = self::getFeSpace(0);
  for (size_t i = 0; i < self::dow; ++i) {
    /// < -K*grad(P) , grad(psi) >
    Operator *opLaplace2 = new Operator(feSpace, feSpace);
    opLaplace2->setUhOld(self::prob->getStageSolution(i)); // P#_i
    addSOT(opLaplace2, constant(-K));
    
    self::prob->addMatrixOperator(*opLaplace2, i+self::dow, i, &minus1, &minus1);
    self::prob->addVectorOperator(*opLaplace2, i+self::dow);
  }
}


template<typename P> 
void PolarizationField_RB<P>::closeTimestep(AdaptInfo *adaptInfo)
{ FUNCNAME("PolarizationField_RB::closeTimestep()");

  calcVectorField();  
  super::closeTimestep(adaptInfo);
}


template<typename P> 
void PolarizationField_RB<P>::writeFiles(AdaptInfo *adaptInfo, bool force)
{ FUNCNAME("PolarizationField_RB::closeTimestep()");

  super::writeFiles(adaptInfo, force);
  self::fileWriter->writeFiles(adaptInfo, false);
}

} // end namespace detail

} }