/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/

#ifndef CAHN_HILLIARD_NAVIER_STOKES_H
#define CAHN_HILLIARD_NAVIER_STOKES_H

#include "AMDiS.h"

// coupling structures
#if __cplusplus > 199711L
  #include "CouplingBaseProblem2_cxx11.h"
#else
  #include "CouplingBaseProblem2.h"
#endif

#include "CahnHilliard_.h"

// structures for local refinement
#include "Refinement.h"
#include "MeshFunction_Level.h"

#include "POperators.h"

using namespace AMDiS;

/**
  * \ingroup Problem 
  *
  * \brief
  */
template<typename CH_Type = CahnHilliard_,
	 typename NS_Type = NavierStokes_TaylorHood>
class CahnHilliardNavierStokes_ : public CouplingBaseProblem<ProblemStat, CH_Type, NS_Type>
{
public:
  
  typedef CouplingBaseProblem<ProblemStat, CH_Type, NS_Type> super;
  
public:

  CahnHilliardNavierStokes_(std::string name_, CH_Type *chProb_, NS_Type *nsProb_)
  : super(name_, *chProb_, *nsProb_),
    chProb(chProb_),
    nsProb(nsProb_),
    refFunction(NULL),
    refinement(NULL),
    sigma(0.0),
    surfaceTension(0.0),
    y(NULL),
    yc1(NULL),
    yc2(NULL)
  {
    Parameters::get(super::name + "->sigma", sigma);
    surfaceTension = sigma/chProb->getEpsilon() * 3.0/2.0/sqrt(2.0);
  }


  ~CahnHilliardNavierStokes_() 
  {
    if (refFunction) {
      delete refFunction;
      refFunction = NULL;
    }
    if (refinement) {
      delete refinement;
      refinement = NULL;
    }
    if (y) {
      delete y;
      y = NULL;
    }
    if (yc1) {
      delete yc1;
      yc1 = NULL;
    }
    if (yc2) {
      delete yc2;
      yc2 = NULL;
    }
  }

  
  /**
   * All needed DOFVectors are initialized so that they can be used inside
   * of the operators.
   **/
  void initData()
  {
    y = new DOFVector<double>(chProb->getFeSpace(0), "volume");
    yc1 = new DOFVector<double>(chProb->getFeSpace(0), "volume");
    yc2 = new DOFVector<double>(chProb->getFeSpace(0), "volume");
    
    output = fopen( "output_data", "w" );
    fprintf(output, "time u_b center_old\n");
    fclose(output);
    u_b = 0;
    
    
    // meshFunction for k local refinement around the interface of the phasefield-DOFVector
    refFunction = new PhaseFieldRefinement(chProb->getMesh());

    if (chProb->getDoubleWellType() == 0) {
      refinement = new RefinementLevelDOF(
	chProb->getFeSpace(),
	refFunction,
	chProb->getSolution()->getDOFVector(1)); // phaseField-DOFVector in [0,1]
    } else {
      refinement = new RefinementLevelDOF(
	chProb->getFeSpace(),
	refFunction,
	new PhaseDOFView<double>(chProb->getSolution()->getDOFVector(1))); // phaseField-DOFVector in [-1,1]
    }
  }
  

  /**
   * In this method the operators responsible for the coupling of all problems are defined
   * and added to the problem. The access to the problems is through the baseBroblem->getProblem()
   * method. All baseProblem provide FiniteElemSpaces and solution vectors.
   **/
  void fillCouplingOperators()
  { FUNCNAME("CahnHilliardNavierStokes::fillCouplingOperators()");
    MSG("CahnHilliardNavierStokes::fillCouplingOperators()");
    
    int degree = nsProb->getFeSpace(0)->getBasisFcts()->getDegree();

    for (size_t i = 0; i < super::dow; i++) {
      // < nu * d_i(c) , theta >
      Operator *opNuGradC = new Operator(nsProb->getFeSpace(i), chProb->getFeSpace(0));
      opNuGradC->addTerm(new VecAndPartialDerivative_ZOT(chProb->getSolution()->getDOFVector(0), 
							 chProb->getSolution()->getDOFVector(1), i));
      nsProb->getProblem(0)->addVectorOperator(opNuGradC, i, &surfaceTension, &surfaceTension);
    }

     for (size_t i = 0; i < super::dow; i++) {
	// < v * grad(c) , theta >
	Operator *opVGradC = new Operator(chProb->getFeSpace(0), chProb->getFeSpace(1));
	opVGradC->addTerm(new VecAndPartialDerivative_FOT(nsProb->getSolution()->getDOFVector(i), i), GRD_PHI);
	chProb->getProblem()->addMatrixOperator(opVGradC, 1, 1,  chProb->getTau());
     }
   /**/ 
  }
  

  /**
   * Solves the initial problems.Before this is done the mesh is refined by the Refinement-class
   * This provides local refinement functions for phaseField-refinement or signed-distance-refinement.
   * See MeshFunction_Level.h for some details.
   **/ 
  void solveInitialProblem(AdaptInfo *adaptInfo)
  {
    // initial refinement
    refinement->refine(0);

    for (int i = 0; i < 5; i++) {
      chProb->solveInitialProblem(adaptInfo); 	// update phaseField-DOFVector
      refinement->refine(5);			// do k steps of local refinement
    }

    // solve all initial problems of the problems added to the CouplingTimeInterface
    CouplingTimeInterface::solveInitialProblem(adaptInfo);
  }


  /**
   * Called at the end of each timestep.
   * If the phase-field has changed the mesh is updated by the refinement-method
   **/
  void closeTimestep(AdaptInfo *adaptInfo)
  {
    CouplingTimeInterface::closeTimestep(adaptInfo);
    
    AMDiS::Component<1> yFun(y->getFeSpace()->getBasisFcts()->getDegree());
    y->interpol(&yFun);

    DOFIterator<double> phiIt(chProb->getSolution()->getDOFVector(1), USED_DOFS);
    DOFIterator<double> yIt(y, USED_DOFS);
    DOFIterator<double> yc1It(yc1, USED_DOFS);
    DOFIterator<double> yc2It(yc2, USED_DOFS);

    for (phiIt.reset(), yIt.reset(),yc1It.reset(),yc2It.reset();
         !phiIt.end();
         ++phiIt, ++yIt,++yc1It,++yc2It){
                *yc2It = ((1.0)+*phiIt)/2.0;
                *yc1It = *yc2It * *yIt;
	}

    if (adaptInfo->getTimestepNumber() <= 1) 
      center_old = (yc1->Int()) / (yc2->Int());
    u_b = - (center_old - (yc1->Int()) / (yc2->Int()) ) / adaptInfo->getTimestep();
    center_old = (yc1->Int()) / (yc2->Int());
    
    // append values to file
    output = fopen( "output_data", "a" );
    fprintf(output, "%f, %f, %f\n", adaptInfo->getTime(), u_b, center_old);
    fclose(output);
    
    refinement->refine(2);
  }

protected:

  CH_Type *chProb;	// CahnHilliard baseProblem
  NS_Type *nsProb;	// Navier-Stokes baseProblem

  PhaseFieldRefinement* refFunction;
  RefinementLevelDOF *refinement;

  double sigma;		// coupling parameter to calculate the surface tension
  double surfaceTension;// := sigma/epsilon
  double u_b, center_old;
  
  FILE *output;
  
  DOFVector<double> *y, *yc1, *yc2;
};

#endif // CAHN_HILLIARD_NAVIER_STOKES_H
