/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/
#include "AMDiS.h"
#include "CahnHilliard_.h"
#include "NavierStokes_TH_MultiPhase.h"
#include "CahnHilliardNavierStokes_.h"
#include "SignedDistFunctors.h"
#include "PhaseFieldConvert.h"

#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
  #include "parallel/PetscSolverCahnHilliard.h"
  #include "parallel/PetscSolverCahnHilliard2.h"
  #include "parallel/PetscSolverNavierStokes.h"
#endif

using namespace AMDiS;

class CahnHilliardPC : public CahnHilliard_
{
public:
  CahnHilliardPC(std::string name) : CahnHilliard_(name), delta(1.0)
  { }
 
  void initData()
  {   
    #ifdef HAVE_PARALLEL_DOMAIN_AMDIS
    string initFileStr = name + "->space->solver", solverType = "";
    Parameters::get(initFileStr, solverType);
    if (solverType == "petsc-ch") {
      Parallel::PetscSolverCahnHilliard* solver = dynamic_cast<Parallel::PetscSolverCahnHilliard*>(prob->getSolver());
      if (solver)
	solver->setChData(&eps, &delta, prob->getSolution());	
    } else if (solverType == "petsc-ch2") {
      Parallel::PetscSolverCahnHilliard2* solver = dynamic_cast<Parallel::PetscSolverCahnHilliard2*>(prob->getSolver()); 
      if (solver)
	solver->setChData(&eps, &delta);
    }
    #endif
  }
  
  virtual void setTime(AdaptInfo* adaptInfo) 
  {
    super::setTime(adaptInfo);
    delta = gamma*adaptInfo->getTimestep();
  }
  
  double delta;
};





class NavierStokes_TaylorHoodPC : public NavierStokes_TH_MultiPhase
{
public:
  typedef NavierStokes_TH_MultiPhase super;
  
public:
  NavierStokes_TaylorHoodPC(std::string name) : super(name)
  { }
 
 
 void initData()
  {   
    super::initData();
#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
    string initFileStr = name + "->space->solver", solverType = "";
    Parameters::get(initFileStr, solverType);
    if (solverType == "petsc-navierstokes") {
      Parallel::PetscSolverNavierStokes *s = dynamic_cast<Parallel::PetscSolverNavierStokes*>(prob->getSolver());
      if (s) {
	  s->setStokesData( getInvTau(), prob->getSolution(), &viscosity1, &viscosity2, &density1, &density2);
	  s->setPhase(multiPhase);
      }
    } 
#endif    
  }
  
  void initPhasefield()
  {   
#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
    string initFileStr = name + "->space->solver", solverType = "";
    Parameters::get(initFileStr, solverType);
    if (solverType == "petsc-navierstokes") {
      Parallel::PetscSolverNavierStokes *s = dynamic_cast<Parallel::PetscSolverNavierStokes*>(prob->getSolver());
      if (s) {
	  s->setPhase(multiPhase);
      }
    } 
#endif    
  }
  
};




class MyNavierStokes_PC : public NavierStokes_TaylorHoodPC
{
public:
  MyNavierStokes_PC(std::string name_) : NavierStokes_TaylorHoodPC(name_) {}
  
  void fillBoundaryConditions()
  { FUNCNAME("MyNavierStokes_PC::fillBoundaryConditions()");
    
    DOFVector<double> *zeroDOF = new DOFVector<double>(getFeSpace(0), "zero");
    zeroDOF->set(0.0);
    size_t dow = Global::getGeo(WORLD);

    /// at rigid wall: no-slip boundary condition  
    // oben-unten
    getProblem()->addDirichletBC(2, 0, 0, zeroDOF);
    getProblem()->addDirichletBC(2, 1, 1, zeroDOF);
    // links-rechts
    getProblem()->addDirichletBC(1, 0, 0, zeroDOF); 
  }
  
  
};


class MyNavierStokes: public NavierStokes_TH_MultiPhase
{
public:
  MyNavierStokes(std::string name_) : NavierStokes_TH_MultiPhase(name_) {}
  
  void fillBoundaryConditions()
  { FUNCNAME("MyNavierStokes::fillBoundaryConditions()");
    
    DOFVector<double> *zeroDOF = new DOFVector<double>(getFeSpace(0), "zero");
    zeroDOF->set(0.0);
    size_t dow = Global::getGeo(WORLD);

    /// at rigid wall: no-slip boundary condition  
    // oben-unten
    getProblem()->addDirichletBC(2, 0, 0, zeroDOF);
    getProblem()->addDirichletBC(2, 1, 1, zeroDOF);
    // links-rechts
    getProblem()->addDirichletBC(1, 0, 0, zeroDOF); 
  }
  
  
};



int main(int argc, char** argv)
{ FUNCNAME("main");

  AMDiS::init(argc, argv);
  Timer t;
  
  #ifdef HAVE_PARALLEL_DOMAIN_AMDIS
      CreatorMap<LinearSolver>::addCreator("petsc-ch", new Parallel::PetscSolverCahnHilliard::Creator);
      CreatorMap<LinearSolver>::addCreator("petsc-ch2", new Parallel::PetscSolverCahnHilliard2::Creator);
  #endif

  #ifdef HAVE_PARALLEL_DOMAIN_AMDIS      
    CahnHilliardPC chProb("ch");
    MyNavierStokes_PC nsProb("ns");
    CahnHilliardNavierStokes_<CahnHilliardPC,MyNavierStokes_PC> mainProb("main", &chProb, &nsProb);
  #else
    CahnHilliard_ chProb("ch");
    MyNavierStokes nsProb("ns");
    CahnHilliardNavierStokes_<CahnHilliard_,MyNavierStokes> mainProb("main", &chProb, &nsProb);
  #endif
  
  mainProb.initialize(INIT_ALL);
  mainProb.initTimeInterface();
  
  
    
  // Adapt-Infos
  AdaptInfo adaptInfo("adapt", mainProb.getNumComponents());
  
  nsProb.setMultiPhase(chProb.getSolution()->getDOFVector(1));
  #ifdef HAVE_PARALLEL_DOMAIN_AMDIS      
    nsProb.initPhasefield();
  #endif
  // adaption loop - solve ch-prob and ns-prob
  AdaptInstationary adaptInstat("adapt", mainProb, adaptInfo, mainProb, adaptInfo);

  int error_code = adaptInstat.adapt(); 

  MSG("elapsed time= %f sec\n", t.elapsed());

  AMDiS::finalize();

  return error_code;
};
