/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/

#include "AMDiS.h"
#include "NavierStokes_TaylorHood.h"

using namespace AMDiS;


struct DrivenCavityBC : AbstractFunction<double, WorldVector<double> >
{
  double operator()(const WorldVector<double> &x) const 
  {
    double vel = std::max(0.0, 1.0 - 4.0 * sqr(x[0] - 0.5));
    return (*time < 0.1 ? sin((*time) * m_pi/0.2)*vel : vel);
  }
  
  void setTimePtr(double* time_)
  {
    time = time_;
  }
  
private:
  double* time;
};

class DrivenCavity : public NavierStokes_TaylorHood
{
public: // typedefs
  
  typedef NavierStokes_TaylorHood super;
  
public: // methods

  /// Constructor
  DrivenCavity(std::string name) : super(name), drivenCavityBC(NULL) { }
  
  /// Destructor
  ~DrivenCavity()
  {
    if (drivenCavityBC) {
      delete drivenCavityBC;
      drivenCavityBC = NULL;
    }
  }
  
  virtual void initData()
  {
    super::initData();
#ifdef HAVE_PARALLEL_DOMAIN_AMDIS
    Parallel::PetscSolverNavierStokes* solver = dynamic_cast<Parallel::PetscSolverNavierStokes*>(prob->getSolver());
    if (solver)
      solver->setStokesData(&viscosity, getInvTau(), prob->getSolution());
      // solver->setStokesData(getInvTau(), prob->getSolution(), &viscosity1, &viscosity2, &density1, &density2);
#endif
  }

  /// Set initial condition: constant zero velocity
  virtual void solveInitialProblem(AdaptInfo *adaptInfo)
  {
    super::solveInitialProblem(adaptInfo);
    
    for (size_t i = 0; i < dow; i++)
      prob->getSolution()->getDOFVector(i)->set(0.0);
    
    drivenCavityBC->setTimePtr(adaptInfo->getTimePtr());
  }
   
  /// add boundary conditions corresponding to the lid driven cavity example
  virtual void fillBoundaryConditions()
  {    
    AMDiS::Const<double, WorldVector<double> > *zero = new AMDiS::Const<double, WorldVector<double> >(0.0);

    /// at rigid wall: no-slip boundary condition
    for (size_t i = 0; i < dow; i++)
      prob->addDirichletBC(1, i, i, zero);

    /// at upper wall: prescribed velocity
    drivenCavityBC = new DrivenCavityBC;
    prob->addDirichletBC(2, 0, 0, drivenCavityBC);
    prob->addDirichletBC(2, 1, 1, zero);
  }

protected: // variables

  DrivenCavityBC *drivenCavityBC;
};


int main(int argc, char** argv)
{ FUNCNAME("main");

  AMDiS::init(argc, argv);  
  Timer t;

  DrivenCavity nsProb("ns");
  nsProb.initialize(INIT_ALL);
  nsProb.initTimeInterface();

  // Adapt-Infos
  AdaptInfo adaptInfo("adapt", nsProb.getNumComponents());
  AdaptInstationary adaptInstat("adapt", nsProb, adaptInfo, nsProb, adaptInfo);

  int error_code = adaptInstat.adapt(); 

  MSG("elapsed time= %d sec\n", t.elapsed());
  AMDiS::finalize();

  return error_code;
}
