/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/
 
#ifndef MTL_CH_PRECONDITIONER_H
#define MTL_CH_PRECONDITIONER_H

#include "AMDiS.h"
#include "solver/BlockPreconditioner.h"

namespace AMDiS { namespace extensions {

/// Block preconditioner for the Cahn-Hilliard equation
/**
 * Based on the paper "Efficient preconditioners for large scale binary 
 * Cahn-Hilliard models", P.Boyanova, M.Do-Quang, M.Neytcheva
 * 
 * The Cahn-Hilliard equation must be provided in the form:
 *
 *    mu - Phi'(c) + eps^2*laplace(c) = 0
 *    -gamma*tau*laplace(mu) + c + tau*(u*nabla)c = c^old
 *
 * respective in matrix form:
 *
 *    [ M          ,  -(J + eps^2*L) ] [mu] = [Phi'(c^old) - J*c^old]
 *    [ gamma*tau*L,  M + tau*W      ] [c ] = [M*c^old              ]
 *
 * with M...mass matrix, L...laplace matrix, J...discretization of 
 * jacobian of Phi', W...discretization of transport term
 *
 */
template<typename MatrixType>
struct MTLPreconChBase : BlockPreconditioner<MatrixType>
{
  typedef MTLTypes::MTLVector              VectorType;
  
  typedef BlockPreconditioner<MatrixType>  super;
  typedef MTLPreconChBase<MatrixType>      self;
  typedef typename super::precon_base      precon_base;
 
  struct Creator : public CreatorInterfaceName<precon_base>
  {
    virtual ~Creator() {}    
    precon_base* create() { return new self(this->name); }
  };
  
  /// constructor
  MTLPreconChBase(std::string name)
    : name(name), 
      eps(0.1), gamma(1.0), tau(NULL),
      solverM(NULL), solverMpL(NULL), runnerM(NULL), runnerMpL(NULL)
  {
    super::createSubSolver(name + "->subsolver M", solverM, runnerM, "cg", "diag", 5, 1.e-3);
    super::createSubSolver(name + "->subsolver MpL", solverMpL, runnerMpL, "cg", "diag", 20, 1.e-3);
  }
  
  
  /// destructor
  virtual ~MTLPreconChBase()
  {
    exit();
    delete solverM;
    delete solverMpL;
  }
  
  /// prepare needed data, e.g. the matrix MpL = M + a*L and initialize the runners
  virtual void init(const SolverMatrix<Matrix<DOFMatrix*> >& A, const MatrixType& fullMatrix);  
  
  /// apply the preconditioner P to a vector b, i.e. solve Px = b
  virtual void solve(const VectorType& b, VectorType& x) const;
  
  virtual void exit()
  {
    runnerM->exit();
    runnerMpL->exit();
  }
  
  const MatrixType& getM() const { return *matM; } // c
  const MatrixType& getL() const { return *matL; } // laplace(c)
  
  /// return delta = \ref tau * \ref gamma
  double getDelta() const { return (*tau) * gamma; }
  
  /// the preconditioner needs some data from the Cahn-Hilliard equation.
  /// This is passed in the struct data.
  template<typename Data>
  void setData(Data* data)
  {
    tau = data->tau;
    eps = data->eps;
    gamma = data->gamma;
    matM = &data->getM();
    matL = &data->getL();
  }
  
protected:      
  /// solve the mass equation Mx = b
  template <typename VectorX, typename VectorB>
  void solveM(VectorX& x, const VectorB& b) const
  {
    runnerM->solve(getM(), x, b);
  }
  
  /// solve a diffusion equation (M+a*L)x = b
  template <typename VectorX, typename VectorB>
  void solveMpL(VectorX& x, const VectorB& b) const
  {
    runnerMpL->solve(MpL, x, b);
  }
  
protected:  
  MatrixType* matM;
  MatrixType* matL;
  
  MatrixType MpL;
   
  mutable VectorType y1;
  mutable VectorType y2;
  
  LinearSolverInterface* solverM;
  LinearSolverInterface* solverMpL;
  
  RunnerBase<MatrixType, VectorType>* runnerM;
  RunnerBase<MatrixType, VectorType>* runnerMpL;
  
  std::string name;
  
  double eps;
  double gamma;
  double* tau;
};


typedef MTLPreconChBase<MTLTypes::MTLMatrix> MTLPreconCahnHilliard;

} }

#include "MTLPreconCahnHilliard.hh"

#endif // MTL_CH_PRECONDITIONER_H

