/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/

namespace AMDiS { namespace extensions { 
  
void MTLPreconCh2Sub::init(const SolverMatrix<Matrix<DOFMatrix*> >& A_, const MatrixType& fullMatrix_)
{ FUNCNAME("MTLPreconCh2Sub::init()");

  super::init(A_, fullMatrix_);
  assert(tau != NULL);  
  
  // helper-matrix MpL = M + eps*sqrt(delta)*L
  MpL.change_dim(num_rows(getM()), num_cols(getM()));
  MpL = getM() + ( eps*std::sqrt(getDelta()) ) * getL();
  
  y1.change_dim(num_rows(getM()));
  y2.change_dim(num_rows(getM()));
  
  runnerMpL->init(A_, MpL);
}


void MTLPreconCh2Sub::solve(const VectorType& b, VectorType& x) const
{ FUNCNAME("MTLPreconCh2Sub::solve()");

  x.change_dim(num_rows(b));
  
  solveMpL(y1, b);
  y2 = getM() * y1; 
  solveMpL(x, y2);
}

  
template< typename MatrixType >
void MTLPreconCh2Base< MatrixType >::init(const SolverMatrix<Matrix<DOFMatrix*> >& A_, const MatrixType& fullMatrix_)
{ FUNCNAME("MTLPreconCh2Base::init()");

  super::init(A_, fullMatrix_);
  assert(tau != NULL);
  
  mtl::dense_vector<double> diag_M = mtl::matrix::diagonal(getM());
  for (size_t i = 0; i < size(diag_M); i++)
    diag_M[i] = 1.0 / diag_M[i];
  MInv = mtl::vector::diagonal(diag_M);
  
  S.change_dim(num_rows(getM()), num_cols(getM()));
  S = getM() + ( sqr(eps)*getDelta() ) * getL() * MInv * getL();
  
  PId = new itl::pc::identity<MatrixType, double>(getM());
  
  // temporary variables
  y1.change_dim(num_rows(getM()));
  y2.change_dim(num_rows(getM()));
  
  runnerM->init(A_, getM());
  preconS.init(A_, S);
}


template< typename MatrixType >
void MTLPreconCh2Base< MatrixType >::solve(const VectorType& b, VectorType& x) const
{ FUNCNAME("MTLPreconCh2Base::solve()");

  x.change_dim(num_rows(b));
  
  const VectorType b1(b[super::getRowRange(0)]);
  const VectorType b2(b[super::getRowRange(1)]);
  
  VectorType x1(x[super::getRowRange(0)]);
  VectorType x2(x[super::getRowRange(1)]);
  
  solveM(y1, b1);
  y2 = getL() * y1; 
  x1 = b2 - getDelta() * y2;
  
  solveS(x2, x1);
  y1 = getL() * x2;
  y2 = b1 + sqr(eps) * y1;

  solveM(x1, y2);
}

} }