/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/
#ifndef MTL_NS_PRECONDITIONER_H
#define MTL_NS_PRECONDITIONER_H

#include "AMDiS.h"
#include "BlockPreconditioner.h"

namespace AMDiS { namespace extensions { 

struct MTLPreconNavierStokes : BlockPreconditioner<MTLTypes::MTLMatrix>
{
  typedef MTLTypes::MTLMatrix MatrixType;
  typedef MTLTypes::MTLVector VectorType;
  
  typedef BlockPreconditioner<MatrixType>                  super;
  typedef MTLPreconNavierStokes                            self;
  typedef ITL_BasePreconditioner<MatrixType, VectorType>   precon_base;
 
  class Creator : public CreatorInterfaceName<precon_base>
  {
  public:
    virtual ~Creator() {}
    
    precon_base* create() { 
      return new self(this->name);
    }
  };  
  
  MTLPreconNavierStokes(std::string name)
    : super(),
      name(name), 
      tau(NULL),
      pressureComponent(Global::getGeo(WORLD)),
      velocitySolutionMode(""),
      massSolutionMode(""),
      laplaceSolutionMode(""),
      massMatrix(NULL), laplaceMatrix(NULL), conDifMatrix(NULL)
      massMatrix_(1,1), laplaceMatrix_(1,1), 
      velocityMatrix_(Global::getGeo(WORLD), Global::getGeo(WORLD)),
      divMatrix_(1, Global::getGeo(WORLD)),
      gradMatrix_(Global::getGeo(WORLD), 1),
      solver_velocity(NULL), runner_velocity(NULL), 
      solver_mass(NULL), runner_mass(NULL),
      solver_laplace(NULL), runner_laplace(NULL)
  {
  
    Parameters::get(name + "->navierstokes->pressure component", 
		    pressureComponent);
		    
    std::string initFileStr_velocity = name + "navierstokes->velocity solver";
    Parameters::get(initFileStr_velocity, velocitySolutionMode);
    
    std::string initFileStr_mass = name + "navierstokes->mass solver";
    Parameters::get(initFileStr_mass, massSolutionMode);
    
    std::string initFileStr_laplace = name + "navierstokes->laplace solver";
    Parameters::get(initFileStr_laplace, laplaceSolutionMode);
		    
    // === read solver-name ===
    std::string solverType_velocity = "mtl_" + velocitySolutionMode;
    std::string solverType_mass = "mtl_" + massSolutionMode;
    std::string solverType_laplace = "mtl_" + laplaceSolutionMode;
  
    // velocity solver
    LinearSolverCreator *solverCreator_velocity = 
      dynamic_cast<LinearSolverCreator*>(CreatorMap<LinearSolver>::getCreator(solverType_velocity, initFileStr_velocity));
    TEST_EXIT(solverCreator_velocity)
      ("No valid solver type found in parameter \"%s\"\n", initFileStr_velocity.c_str());
    solverCreator_velocity->setName(initFileStr_velocity);
    solver_velocity = solverCreator_velocity->create();
    runner_velocity = dynamic_cast<RunnerBase< MatrixType, VectorType >*>(solver_velocity->getRunner());
    
    // mass solver
    LinearSolverCreator *solverCreator_mass = 
      dynamic_cast<LinearSolverCreator*>(CreatorMap<LinearSolver>::getCreator(solverType_mass, initFileStr_mass));
    TEST_EXIT(solverCreator_mass)
      ("No valid solver type found in parameter \"%s\"\n", initFileStr_mass.c_str());
    solverCreator_mass->setName(initFileStr_mass);
    solver_mass = solverCreator_mass->create();
    runner_mass = dynamic_cast<RunnerBase< MatrixType, VectorType >*>(solver_mass->getRunner());
    
    // laplace solver
    LinearSolverCreator *solverCreator_laplace = 
      dynamic_cast<LinearSolverCreator*>(CreatorMap<LinearSolver>::getCreator(solverType_laplace, initFileStr_laplace));
    TEST_EXIT(solverCreator_laplace)
      ("No valid solver type found in parameter \"%s\"\n", initFileStr_laplace.c_str());
    solverCreator_laplace->setName(initFileStr_laplace);
    solver_laplace = solverCreator_laplace->create();
    runner_laplace = dynamic_cast<RunnerBase< MatrixType, VectorType >*>(solver_laplace->getRunner());
  }
  
  virtual ~MTLPreconNavierStokes()
  {
    exit();
  }
  
  void setStokesData(double *invTauPtr, SystemVector *vec, double *nu1_=NULL, double *nu2_=NULL, double *rho1_=NULL, double *rho2_=NULL)
  {
    nu = new double;	
    (*nu) = 0.0;
    invTau = invTauPtr;
    solution = vec;
    nu1=nu1_;
    nu2=nu2_;
    rho1=rho1_;
    rho2=rho2_;      
  }
  
  void setStokesData(double *nuPtr, double *invTauPtr, SystemVector *vec)
  {
    nu = nuPtr;
    invTau = invTauPtr;
    solution = vec;
  }
  
  
  void setPhase(DOFVector<double> *d)
  {
    phase = d;
  }
  
  virtual void init(const typename super::BlockMatrix& A, const MatrixType& fullMatrix);
  
  virtual void exit()
  {    
    runner_laplace->exit();
    runner_mass->exit();
    runner_velocity->exit();
  
    if (massMatrix) { delete massMatrix; massMatrix = NULL; }
    if (laplaceMatrix) { delete laplaceMatrix; laplaceMatrix = NULL; }
    if (conDifMatrix) { delete conDifMatrix; conDifMatrix = NULL; }
    if (solver_velocity) { delete solver_velocity; solver_velocity = NULL; }
    if (solver_mass) { delete solver_mass; solver_mass = NULL; }
    if (solver_laplace) { delete solver_laplace; solver_laplace = NULL; }
  }
  
  virtual void solve(const MTLVector& b, MTLVector& x) const;  
  
  const MatrixType& getM() const { return massMatrix->getBaseMatrix(); }
  const MatrixType& getL() const { return laplaceMatrix->getBaseMatrix(); }
  const MatrixType& getConDif() const { return conDifMatrix->getBaseMatrix(); }
  const MatrixType& getMatV() const { return blockVelocityMatrix->getMatrix(); }
  
  const MatrixType& getMatDiv() const { return blockDivMatrix->getMatrix(); }
  const MatrixType& getMatGrad() const { return blockGradMatrix->getMatrix(); }
  
protected: 
  DOFMatrix* massMatrix;
  DOFMatrix* laplaceMatrix;
  DOFMatrix* conDifMatrix;
  
  mtl::irange vel_rows;
  
  Matrix<DOFMatrix*> massMatrix_;
  SolverMatrix<Matrix<DOFMatrix*> > blockMassMatrix; 
  
  Matrix<DOFMatrix*> laplaceMatrix_;
  SolverMatrix<Matrix<DOFMatrix*> > blockLaplaceMatrix; 
  
  Matrix<DOFMatrix*> velocityMatrix_;
  SolverMatrix<Matrix<DOFMatrix*> > blockVelocityMatrix; 
  
  Matrix<DOFMatrix*> divMatrix_;
  SolverMatrix<Matrix<DOFMatrix*> > blockDivMatrix; 
  
  Matrix<DOFMatrix*> gradMatrix_;
  SolverMatrix<Matrix<DOFMatrix*> > blockGradMatrix; 
   
  mutable VectorType y01;
  mutable VectorType y2;
  
  int pressureComponent;
  std::string velocitySolutionMode;
  std::string massSolutionMode;
  std::string laplaceSolutionMode;
    
  LinearSolverInterface* solver_velocity;
  LinearSolverInterface* solver_mass;
  LinearSolverInterface* solver_laplace;
  
  RunnerBase< MatrixType, VectorType >* runner_velocity;
  RunnerBase< MatrixType, VectorType >* runner_mass;
  RunnerBase< MatrixType, VectorType >* runner_laplace;

  std::string name;
  
  double *nu, *invTau, *nu1,*nu2,*rho1,*rho2;
};

} }

#include "MTLPreconNavierStokes.hh"

#endif // MTL_NS_PRECONDITIONER_H

