/******************************************************************************
 *
 * Extension of AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: Simon Praetorius et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/
#include "PetscPreconCahnHilliard.h"

namespace AMDiS {

  using namespace std;
  
  /// solve cahn-hilliard Preconditioner
  PetscErrorCode pcChShell(PC pc, Vec b, Vec x) // solve Px=b
  { 
    void *ctx;
    PCShellGetContext(pc, &ctx);
    ChData* data = static_cast<ChData*>(ctx);
    
    Vec b1, b2, x1, x2;
    VecNestGetSubVec(b, 0, &b1);
    VecNestGetSubVec(b, 1, &b2);

    VecNestGetSubVec(x, 0, &x1);
    VecNestGetSubVec(x, 1, &x2);

    // Hilfsvariablen
    Vec y1, y2;
    VecDuplicate(b1, &y1);
    VecDuplicate(b1, &y2);

    KSPSolve(data->kspM, b1, y1); 			// M*y1 = b1    
    MatMultAdd(data->matMinusDeltaK, y1, b2, x1); 	// -> x1 := b2-delta*K*y1
    
    KSPSolve(data->kspMpK, x1, y2); 			// (M+eps*sqrt(delta))*y2 = x1
    MatMult(data->matM, y2, x1); 			// x1 := M*y2
    
    KSPSolve(data->kspMpK, x1, x2);			// (M+eps*sqrt(delta))*x2 = x1
    double factor = (data->eps)/sqrt(data->delta);
    VecCopy(x2, x1); 					// x1 := x2
    VecAXPBYPCZ(x1, 1.0, factor, -factor, y1, y2);	// x1 = 1*y1 + factor*y2 - factor*x1
    
    VecDestroy(&y1);
    VecDestroy(&y2);
    
    PetscFunctionReturn(0);
  }
  

  void PetscPreconCahnHilliard::init(PC pc, const SolverMatrix<Matrix<DOFMatrix*> >&, const MatrixType& A)
  {    
    PCSetType(pc, PCSHELL);
    PCShellSetApply(pc, pcChShell);
    PCShellSetContext(pc, &data);
    
    MatNestGetSubMat(A.matrix, 1, 0, &data.matM);
    MatDuplicate(data.matM, MAT_COPY_VALUES, &data.matMinusDeltaK);
    MatScale(data.matMinusDeltaK, -1.0);
    
    MatNestGetSubMat(A.matrix, 0, 0, &data.matM);
    
    MatDuplicate(data.matM, MAT_COPY_VALUES, &MpK);
    MatAXPY(MpK, -(*eps)/sqrt(*delta), data.matMinusDeltaK, SAME_NONZERO_PATTERN);
    
    Runner::createSubSolver(data.kspM, data.matM, "mass_");
    Runner::createSubSolver(data.kspMpK, MpK, "MpK_");

    Runner::setSolver(data.kspM, "mass_", KSPCG, PCJACOBI, 0.0, 1e-14, 15);
    Runner::setSolver(data.kspMpK, "MpK_", KSPRICHARDSON, PCHYPRE, 0.0, 1e-14, 15);
    
    // === Setup preconditioner data ===    
    data.delta = *delta;
    data.eps = *eps;
  }


  void PetscPreconCahnHilliard::exit()
  {
    MatDestroy(&data.matMinusDeltaK);
    MatDestroy(&MpK);
    
    KSPDestroy(&data.kspM);
    KSPDestroy(&data.kspMpK);
  }
}
